/*
 * Copyright (c) 2012 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                      Anthony Mallet on Thu Mar 15 2012
 */
#ifndef H_CCLIENT_PRIVATE
#define H_CCLIENT_PRIVATE

#include "ros/node_handle.h"

#include "genom3/c/client.h"

#define GENOM_COMP_ERR_FLAG	(10000U<<16)
#define GENOM_CLIENT_ERR_FLAG	(0xffffU<<16)
#define GENOM_SYS_ERR_FLAG	(0x7fffU<<16)

#define ROS_CLIENT_FREE		0
#define ROS_CLIENT_WAIT_ACK	1
#define ROS_CLIENT_WAIT_TERM	2
#define ROS_CLIENT_TERM		3

struct ros_options {
  char *instance;
};

struct connections;

struct genom_client_s {
  char *instance;
  int error;
  ros::NodeHandle *node;
  struct connections *connections;

  struct genom_client_request {
    const struct genom_service_info *info;
    int aid;
    int status, pstatus;
    int error;
    void *output;
    genom_request_cb sentcb, donecb;
    void *cb_data;
  } *active_rqst;
  int nactive, next;
};

int	ros_init(int argc, char *argv[], char ebuffer[256], size_t esize);
void	ros_fini(void);
int	ros_getopt(int argc, char *argv[], struct ros_options *opts);

int	ros_eventfd(void);
const struct genom_service_info *
	ros_service_info(genom_client h, int rqstid);
int	ros_waitevent(void);
int	ros_waitack(genom_client h, int rqstid);
int	ros_waitterm(genom_client h, int rqstid);
int	ros_done(genom_client h, int rqstid);
int	ros_clean(genom_client h, int rqstid);
int	ros_doevents(genom_client h);

int	ros_newrqst(genom_client h, size_t outsize, int *rqstid);

#endif /* H_CCLIENT_PRIVATE */
