/*
 * Copyright (c) 2012 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *					Anthony Mallet on Fri Feb  3 2012
 */

#ifndef H_GENOM_SERIALIZE
#define H_GENOM_SERIALIZE

#include <err.h>

#include <cstdlib>
#include <cstring>

#include "ros/serialization.h"


/* --- extend ros serializers to basic types ------------------------------- */

namespace ros {
  namespace serialization {
<'foreach type {char double int32_t uint32_t} {'>
    template<size_t N> struct Serializer<<"$type">[N]> {
      template<typename Stream> inline static void
        write(Stream& stream, const <"$type"> (&a)[N]) {
        memcpy(stream.advance(sizeof(a)), a, sizeof(a));
      }

      template<typename Stream> inline static void
        read(Stream& stream, <"$type"> (&a)[N]) {
        memcpy(a, stream.advance(sizeof(a)), sizeof(a));
      }

      inline static uint32_t
        serializedLength(const <"$type"> (&a)[N]) {
        return sizeof(a);
      }
    };
<'}'>

    template<> struct Serializer<char *> {
      template<typename Stream> inline static void
        write(Stream& stream, const char *a) {
        const size_t l = a ? strlen(a) : 0;
        stream.next((uint32_t)l);
        if (l > 0)
          memcpy(stream.advance(l), a, l);
      }

      template<typename Stream> inline static void
        read(Stream& stream, char *(&a)) {
        uint32_t l;
        char *b;
        stream.next(l);

        b = (char *)realloc(a, l+1);
        if (!b) return;
        b[l] = 0;
        a = b;
        if (l == 0) return;
        memcpy(a, stream.advance(l), l);
      }

      inline static uint32_t
        serializedLength(const char *a) {
        return sizeof(uint32_t) + (a ? strlen(a) : 0);
      }
    };
  };
};


#endif /* H_GENOM_SERIALIZE */
