<'
# Copyright (c) 2011-2012 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Tue Nov 29 2011
#
#

# check arguments
if {[llength $argv] != 1} { error "expected arguments: types" }
lassign $argv types

lang c

# compute handy shortcuts
set component [dotgen component]
set comp [$component name]
set COMP [string toupper [$component name]]
'>

#ifndef H_GROS_<"$COMP">_SERIALIZE
#define H_GROS_<"$COMP">_SERIALIZE

#include "ros/serialization.h"

#include "genom-serialize.h"


/* === All in one serializers ============================================== */

<'foreach type $types {
  if {[$type kind] == "typedef"} continue

  # no serialization for basic types
  switch -- [[$type final] kind] {
    {boolean}				-
    {unsigned short} - {short}		-
    {unsigned long} - {long}		-
    {unsigned long long} - {long long}	-
    {float} - {double}			-
    {char} - {octet}			-
    {enum} - {string} - {array} {
      continue
    }

    sequence {
      continue
    }
  }
'>

/* <"[--- Type [$type fullname] --------------------------------------------]"> */

namespace ros {
  namespace serialization {
    template<> struct Serializer< <"[$type declarator]"> > {
      template<typename Stream, typename T>
      inline static void allInOne(Stream& stream, T m) {
<'
  switch -- [$type kind] {
    {struct} - {union} {
      foreach m [$type members] {
'>
        stream.next(m.<"[cname [$m name]]">);
<'
      }
    }

    default {
'>
        stream.next(m);
<'
    }
  }
'>
      }
      ROS_DECLARE_ALLINONE_SERIALIZER;
    };
  };
};

<'}'>

/* === Serialization methods =============================================== */

<'foreach type $types {
  if {[$type kind] == "typedef"} continue

  switch -- [[$type final] kind] {
    {sequence}	{}
    default	{ continue }
  }

  set stype [[$type type] final]
  switch -- [$stype kind] {
    {boolean}				-
    {unsigned short} - {short}		-
    {unsigned long} - {long}		-
    {unsigned long long} - {long long}	-
    {float} - {double}			-
    {char} - {octet}			-
    {enum} {
      set sbasic 1
    }

    default { set sbasic 0 }
  }

'>

/* <"[--- Type [$type declarator] ------------------------------------------]"> */

namespace ros {
  namespace serialization {
    template<> struct Serializer< <"[$type declarator]"> > {
      template<typename Stream> inline static void
        write(Stream& stream, const <"[$type declarator &a]">) {
<'  if {[catch {$type length} l]} {'>
        stream.next(a._length);
<'    if {$sbasic} {'>
        memcpy(stream.advance(a._length * sizeof(*a._buffer)),
               a._buffer, a._length * sizeof(*a._buffer));
<'    } else {'>
        for(uint32_t i = 0; i < a._length; i++) {
          stream.next(a._buffer[i]);
        }
<'    }'>
<'  } else {'>
<'    if {$sbasic} {'>
        memcpy(stream.advance(<"$l"> * sizeof(*a._buffer)),
               a._buffer, <"$l"> * sizeof(*a._buffer));
<'    } else {'>
        for(uint32_t i = 0; i < <"$l">; i++) {
          stream.next(a._buffer[i]);
        }
<'    }'>
<'  }'>
      }

      template<typename Stream> inline static void
        read(Stream& stream, <"[$type declarator &a]">) {
<'  if {[catch {$type length} l]} {'>
        uint32_t l;
        stream.next(l);
        genom_sequence_reserve(&a, l);
        a._length = l;
<'    if {$sbasic} {'>
        memcpy(a._buffer, stream.advance(a._length * sizeof(*a._buffer)),
               a._length * sizeof(*a._buffer));
<'    } else {'>
        for(uint32_t i = 0; i < a._length; i++) {
          stream.next(a._buffer[i]);
        }
<'    }'>
<'  } else {'>
        a._length = <"$l">;
<'    if {$sbasic} {'>
        memcpy(stream.advance(<"$l"> * sizeof(*a._buffer)),
               a._buffer, <"$l"> * sizeof(*a._buffer));
<'    } else {'>
        for(uint32_t i = 0; i < <"$l">; i++) {
          stream.next(a._buffer[i]);
        }
<'    }'>
<'  }'>
      }

      inline static uint32_t
        serializedLength(const <"[$type declarator &a]">) {
<'  if {[catch {$type length} l]} {'>
<'    if {$sbasic} {'>
        return 4 + a._length * sizeof(*a._buffer);
<'    } else {'>
        uint32_t s = 4;
        for(uint32_t i = 0; i < a._length; i++) {
          s += serializationLength(a._buffer[i]);
        }
        return s;
<'    }'>
<'  } else {'>
<'    if {$sbasic} {'>
        return <"$l"> * serializationLength(a._buffer[0]);
<'    } else {'>
        uint32_t l = 0;
        for(uint32_t i = 0; i < <"$l">; i++) {
          l += serializationLength(a._buffer[i]);
        }
        return l;
<'    }'>
<'  }'>
      }
    };
  };
};

<'}'>

#endif /* H_GROS_<"$COMP">_SERIALIZE */
