<'
# Copyright (c) 2011-2012 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Wed Jan 25 2012
#

if {[llength $argv] != 1} { error "expected arguments: component" }
lassign $argv component

# compute handy shortcuts
set comp [$component name]
set COMP [string toupper [$component name]]
if {[catch { $component version } version]} { set version {}}

lang c
'>
#ifndef H_GROS_<"$COMP">_ACTIVITIES
#define H_GROS_<"$COMP">_ACTIVITIES

#include <err.h>
#include <pthread.h>

#include "ros/ros.h"
#include "ros/callback_queue_interface.h"
#include "actionlib/server/action_server.h"


/* --- activities ---------------------------------------------------------- */

enum genom_activity_status {
  ACT_WAIT, ACT_RUN, ACT_STOP, ACT_INTER, ACT_ETHER
};


struct genom_activity : public ros::CallbackInterface {
  int id;
  int rqstid;
  pthread_t task;
  genom_activity_status status;
  <"[[$component event] declarator state]">;

  virtual const actionlib_msgs::GoalID gid(void) = 0;

  virtual int	interrupt(void) = 0;
  virtual <"[[$component event] declarator]"> accept(void) = 0;
  virtual void	reject(<"[[$component event] declarator report]">) = 0;
  virtual void	stop(void) = 0;
  virtual void	report(void) = 0;
  virtual int	invoke(void) = 0;

  void		done(void);
  CallResult	call(void);
};

template<class ActionSpec> struct genom_activity_data : genom_activity {
  typedef typename ActionSpec::_action_goal_type::_goal_type intype;
  typedef typename ActionSpec::_action_result_type::_result_type outype;

  actionlib::ServerGoalHandle<ActionSpec> rqst;
  intype in;
  outype out;

  genom_activity_data(int rqstid_, pthread_t tid,
                      actionlib::ServerGoalHandle<ActionSpec> &rqst_)
    : rqst(rqst_) {
    rqstid = rqstid_;
    task = tid;
    genom::ids::pcopy(in, *rqst_.getGoal());
  };

  const actionlib_msgs::GoalID gid(void) {
    return rqst.getGoalID();
  }

  int interrupt(void);
  <"[[$component event] declarator]"> validate(void);

  <"[[$component event] declarator]"> accept(void) {
     <"[[$component event] declarator s]">;
     s = validate();
     if (s != <"[[[$component event] member ok] declarator]">)
       return s;

     rqst.setAccepted();
     rqst.publishFeedback(id);
     status = ACT_RUN;
     state = <"[[[$component event] member start] declarator]">;
     return s;
   }

   void reject(<"[[$component event] declarator report]">) {
     out.report_ = report;
     rqst.setRejected(out);
     status = ACT_ETHER;
   }

   void stop(void) {
     if (status == ACT_RUN) {
       status = ACT_STOP;
       state = <"[[[$component event] member stop] declarator]">;
     }
   }

   void report() {
     out.report_ = state;

     if (status == ACT_INTER)
       rqst.setCanceled(out);
     else if (status == ACT_ETHER &&
              state == <"[[[$component event] member ok] declarator]">)
       rqst.setSucceeded(out);
     else
       rqst.setAborted(out);
     status = ACT_ETHER;
   }

   int invoke(void);
};

class genom_activities {
  size_t last;

public:
  static const size_t MAX_ACTIVITIES = 32;
  genom_activity *activities[MAX_ACTIVITIES];

  <"[[$component event] declarator]"> alloc(genom_activity *a);
  void free(genom_activity *a);
  genom_activity *bygid(const actionlib_msgs::GoalID &gid);
};

#endif /* H_GROS_<"$COMP">_ACTIVITIES */
