<'
# Copyright (c) 2011-2012 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Fri Dec  9 2011
#

if {[llength $argv] != 1} { error "expected arguments: component" }
lassign $argv component

# compute handy shortcuts
set comp [$component name]
set COMP [string toupper [$component name]]
if {[catch { $component version } version]} { set version {}}

lang c
'>

#include <cstdio>
#include <getopt.h>
#include <libgen.h>
#include <stdlib.h>

#include "ros/ros.h"

#include "<"$comp">_internals.h"

/* options */
static const char usage_string[] = "\
  -b,--background   daemonize\n\
  -i,--name=string  set instance name\n\
     --version      display the version number of the invoked GenoM\n\
  -h,--help         print usage summary (this text)\n\
";
static const char shortopts_string[] = "+bi:h";
static struct option longopts_list[] = {
  { "background",	no_argument,		NULL,	'b'},
  { "name",		required_argument,	NULL,	'i'},
  { "version",		no_argument,		NULL,	-'v'},
  { "help",		no_argument,		NULL,	'h'},
  { NULL, 0, NULL, 0}
};


static void	usage(FILE *channel, char *argv0);


/* --- main ---------------------------------------------------------------- */

/** Initialization and startup
 */
int
main(int argc, char **argv)
{
  int daemonize = 0;
  const char *instance = "<"$comp">";

  int s, c;

  /* parse command line options */
  while (
    (c = getopt_long(argc, argv, shortopts_string, longopts_list, NULL)) != -1)
    switch (c) {
      case 0: break;

      case 'b':
        daemonize++;
	break;

      case 'i':
        instance = optarg;
	break;

      case -'v': puts("<"$version">"); exit(0); break;
      case 'h':  usage(stdout, argv[0]);  exit(0); break;

      case '?':
      default:
	usage(stderr, argv[0]);
        exit(1);
        break;
    }
  argc -= optind;
  argv += optind;

  ros::init(argc, argv, instance);
  s = <"$comp">_cntrl_task(instance);

  return s;
}


/* --- usage --------------------------------------------------------------- */

/** Print usage on a channel.
 *
 * \param[in] channel	Output channel
 * \param[in] argv0	Program invocation name
 */

static void
usage(FILE *channel, char *argv0)
{
  fprintf(channel,
	  "<"[string trim "$comp $version"]"> GenoM component\n\n"
	  "Usage:\n  %1$s [-i name] [-b] [--version]\n"
	  "\n%2$s",
	  basename(argv0), usage_string);
}
